import logging
from datetime import datetime
from typing import Any, Optional

from typeguard import typechecked

from psi_elt.abstract_elt_job import AbstractELTJob, TRANSFORMER_ELT_JOB_TYPE


@typechecked
class CustomTransformer(AbstractELTJob):
    """A CustomTransformer is the third of the ELT jobs that is responsible for
    reading data (from S3, Athena, or Timestream) and transforming the data for
    use in a data mart.

    This class should be treated as abstract and should be extended.
    Child classes must override the _read_data and _transform functions.
    This class includes utilties for reading from S3, Athena, and Timestream.
    """

    def __init__(
        self,
        name: str,
        capture_time: datetime,
        aws_wrangler: Optional[Any] = None,
        timestream_loader: Optional[Any] = None,
        aws_session: Optional[Any] = None,
        logger: Optional[logging.Logger] = None,
    ):
        super().__init__(
            name=name,
            elt_job_type=TRANSFORMER_ELT_JOB_TYPE,
            capture_time=capture_time,
            aws_session=aws_session,
            logger=logger,
        )
        self.aws_wrangler = aws_wrangler
        self.timestream_loader = timestream_loader

    def transform(
        self,
        is_test: bool,
        transform_kwargs: Optional[dict] = None,
        timestream_load_kwargs: Optional[dict] = None,
    ):
        transform_kwargs = self.remove_none_from_dict(transform_kwargs)
        timestream_load_kwargs = self.remove_none_from_dict(timestream_load_kwargs)
        self._transform(is_test, timestream_load_kwargs=timestream_load_kwargs, **transform_kwargs)

    def _transform(
        self, is_test: bool, timestream_load_kwargs: Optional[dict] = None, **transform_kwargs
    ):
        raise Exception("CustomTransformer._transform must be overridden by the child class")

    def write_to_timestream(
        self,
        data: list,
        capture_time: datetime,
        database_name: Optional[str] = None,
        table_name: Optional[str] = None,
        source: Optional[str] = None,
        upsert: bool = False,
        ignore_parse_error: bool = False,
        max_batch_size: Optional[int] = None,
        min_common_dimensions: Optional[int] = None,
        is_multi: bool = True,
        is_preview: bool = False,
    ):
        """Calls TimestreamLoader.load() or load_preview() and raises an
        exception if there are any invalid items or rejected records
        """
        timestream_load_input = {
            "data": data,
            "capture_time": capture_time,
            "database_name": database_name,
            "table_name": table_name,
            "source": source,
            "upsert": upsert,
            "ignore_parse_error": ignore_parse_error,
            "max_batch_size": max_batch_size,
            "min_common_dimensions": min_common_dimensions,
            "is_multi": is_multi,
        }
        timestream_load_input = self.remove_none_from_dict(timestream_load_input)
        if is_preview:
            result = self.timestream_loader.load_preview(**timestream_load_input)
        else:
            result = self.timestream_loader.load(**timestream_load_input)

        if len(result["invalid_items"]) > 0:
            raise Exception(
                f"Failed to write to Timestream with invalid items {result['invalid_items']}"
            )
        if len(result["rejected_records"]) > 0:
            raise Exception(
                f"Failed to write to Timestream with rejected records {result['rejected_records']}"
            )



import json
import re
import logging
from datetime import datetime, time
from typing import Any, Optional

import boto3
import pytz
from typeguard import typechecked

from psi_elt import get_logger

EXTRACTOR_ELT_JOB_TYPE = "extractor"
LOADER_ELT_JOB_TYPE = "loader"
TRANSFORMER_ELT_JOB_TYPE = "transformer"
ELT_JOB_TYPES = [EXTRACTOR_ELT_JOB_TYPE, LOADER_ELT_JOB_TYPE, TRANSFORMER_ELT_JOB_TYPE]


@typechecked
class AbstractELTJob:
    def __init__(
        self,
        name: str,
        elt_job_type: str,
        capture_time: Optional[datetime] = None,
        aws_session: Optional[Any] = None,
        logger: Optional[logging.Logger] = None,
    ):
        assert name, "A name is required to initialize an AbstractELTJob"
        assert elt_job_type in ELT_JOB_TYPES, f"Invalid ELT job type: {elt_job_type}"
        self.__name = name
        self.__elt_job_type = elt_job_type
        self.__aws_session = aws_session or boto3.Session()
        self.__logger = logger or get_logger()

        # Default capture time of 6am UTC
        if capture_time is None:
            self.logger.warning("No capture time provided, using now as the capture time")
            self.__capture_time = datetime.now().astimezone(pytz.utc)
        elif not capture_time.tzinfo:
            self.__capture_time = capture_time.replace(tzinfo=pytz.utc)
        else:
            self.__capture_time = capture_time.astimezone(pytz.utc)
        self.__default_capture_time = datetime.combine(
            self.__capture_time.date(), time(hour=6, tzinfo=pytz.utc)
        )
        self.__default_capture_time_ms = self.dt_to_ms(self.__default_capture_time)
        self.logger.info(f"Initialized ELT job: {self}")

    @property
    def name(self) -> str:
        return self.__name

    @property
    def elt_job_type(self) -> str:
        return self.__elt_job_type

    @property
    def capture_time(self) -> datetime:
        return self.__capture_time

    @property
    def default_capture_time(self) -> datetime:
        return self.__default_capture_time

    @property
    def default_capture_time_ms(self) -> int:
        return self.__default_capture_time_ms

    @property
    def aws_session(self) -> Any:
        return self.__aws_session

    @property
    def logger(self) -> logging.Logger:
        return self.__logger

    def get_s3_object(self, s3_uri: str) -> Any:
        assert s3_uri

        s3_path = s3_uri.split("s3://")[-1]
        bucket = s3_path.split("/", 1)[0]
        key = s3_path.split("/", 1)[-1]

        try:
            client = self.aws_session.client("s3")
            response = client.get_object(Bucket=bucket, Key=key)
            return response["Body"].read().decode("utf-8")
        except Exception as e:
            self.logger.error(f"Failed to get s3 object {s3_path} with exception: {e}")
            raise e

    def get_parameter(self, parameter: str) -> str:
        assert parameter
        try:
            client = self.aws_session.client("ssm")
            response = client.get_parameter(Name=parameter, WithDecryption=True)
            value = response.get("Parameter", {}).get("Value")
            if value is None:
                raise Exception("Parameter value not found in response")
            return value
        except Exception as e:
            self.logger.error(f"Failed to get parameter {parameter} with exception: {e}")
            raise e

    def format_date(self, dt: datetime) -> str:
        """Formats the date as a string. If the provided datetime is naive (no timezone), this
        assumes UTC and returns a string like "2021-05-26T01:02:03Z". If the provided datetime has
        a non-UTC timezone, returns a string like "2021-05-26T01:02:03-05:00".
        """
        if dt.tzinfo and dt.tzinfo != pytz.UTC:
            return dt.isoformat(timespec="seconds")
        return dt.strftime("%Y-%m-%dT%H:%M:%SZ")

    def parse_date(self, dt: str) -> datetime:
        """Parses the given string into a datetime. By default, expects UTC offset information in
        the given string (e.g. "2021-05-26T01:02:03-05:00") and assigns the appropriate time zone.
        If dt is a string like "2021-05-26T01:02:03Z", assumes the timezone is UTC.
        """
        try:
            if dt.endswith("Z"):
                parsed = datetime.strptime(dt, "%Y-%m-%dT%H:%M:%SZ")
                return pytz.timezone("UTC").localize(parsed)
            return datetime.strptime("".join(dt.rsplit(":", 1)), "%Y-%m-%dT%H:%M:%S%z")
        except Exception as e:
            raise ValueError(
                'Datetime must be a string like "2021-05-26T01:02:03Z" for UTC times or "2021-05-26T01:02:03-05:00" for non-UTC times'
            )

    def is_tappid_compliant(self, t_appid: str) -> bool:
        return bool(re.match("^SVC[0-9]{5}$", t_appid))

    def dt_to_ms(self, dt: datetime) -> int:
        if dt:
            return int(round(dt.timestamp() * 1000))
        return -1

    def remove_none_from_dict(self, d: dict) -> dict:
        d = d or {}
        return {k: v for k, v in d.items() if v is not None}

    def to_dict(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "elt_job_type": self.elt_job_type,
            "capture_time": str(self.capture_time),
            "default_capture_time": str(self.default_capture_time),
            "default_capture_time_ms": self.default_capture_time_ms,
        }

    def __repr__(self) -> str:
        return json.dumps(self.to_dict())



import logging
from datetime import datetime
from typing import Any, Optional

from typeguard import typechecked

from psi_elt.abstract_runner import AbstractRunner, NoneType
from psi_elt.custom_transformer import CustomTransformer
from psi_elt.utils import filter_dict


@typechecked
class TransformerRunner(AbstractRunner):
    def __init__(
        self,
        transformers: dict[str, Any],
        aws_wrangler: Optional[Any] = None,
        timestream_loader: Optional[Any] = None,
        logger: Optional[logging.Logger] = None,
    ):
        """
        Initializes a TransformerRunner.

        Accepts a transformers dict, mapping transformer name to transformer class.
        For example:
        {
            "tag-compliance": TagComplianceTransformer,
            "qualys": QualysTransformer
        }
        """
        super().__init__(jobs=transformers, logger=logger)
        self.__aws_wrangler = aws_wrangler
        self.__timestream_loader = timestream_loader

    def parse_event(self, event: dict[str, Any]) -> dict[str, Any]:
        """Parses a transformer event. Child classes may override this function to
        implement custom event parsing rules.

        #### Required event keys
           1. name : the name of the transformer to run

        #### Optional event keys
          1. capture_time                         : target time for the transformer to run against
          2. is_test                              : boolean for whethor or not this is a test run
          3. input_options.athena_workgroup       : Athena workgroup to use when reading data
          4. input_options.athena_data_catalog    : Athena data catalog to use when reading data
          5. input_options.athena_database        : Athena database to read data from
          6. output_options.database_name         : Timestream database to write data to
          7. output_options.table_name            : Timestream table to write data to
          8. output_options.write_to_timestream   : boolean for whether to write to Timestream
          9. output_options.upsert                : boolean for whether to upsert or insert data
         10. output_options.max_batch_size        : int max batch size used when batching
         11. output_options.min_common_dimensions : int min common dimensions used when batching
         12. output_options.is_multi              : boolean for whether to use multi-measure records
        """
        parsed_event = {
            "name": event.get("name"),
            "capture_time": self.get_capture_time(event),
            "job_init_kwargs": {},
            "is_test": bool(event.get("is_test")),
            "transform_kwargs": {},
            "timestream_load_kwargs": {},
        }

        # Parse input options
        parsed_event["transform_kwargs"] = filter_dict(
            dct=event.get("input_options"),
            keys=["athena_workgroup", "athena_data_catalog", "athena_database"],
        )

        # Parse output options
        parsed_event["timestream_load_kwargs"] = filter_dict(
            dct=event.get("output_options"),
            keys=[
                "database_name",
                "table_name",
                "upsert",
                "max_batch_size",
                "min_common_dimensions",
                "is_multi",
            ],
        )
        parsed_event["transform_kwargs"].update(
            filter_dict(dct=event.get("output_options"), keys=["write_to_timestream"])
        )
        return parsed_event

    def validate_parsed_event(self, parsed_event: dict[str, Any]):
        """Validates that a parsed event has the required format"""
        schema = {
            "name": str,
            "capture_time": datetime,
            "job_init_kwargs": {},
            "is_test": bool,
            "transform_kwargs": {
                "athena_workgroup": [str, NoneType],
                "athena_data_catalog": [str, NoneType],
                "athena_database": [str, NoneType],
                "write_to_timestream": [bool, NoneType],
            },
            "timestream_load_kwargs": {
                "database_name": [str, NoneType],
                "table_name": [str, NoneType],
                "upsert": [bool, NoneType],
                "ignore_parse_error": [bool, NoneType],
                "max_batch_size": [int, NoneType],
                "min_common_dimensions": [int, NoneType],
                "is_multi": [bool, NoneType],
            },
        }
        try:
            self._validate_dict(parsed_event, schema)
        except AssertionError as e:
            self.logger.error("Invalid event")
            raise e

    def _run(
        self,
        name: str,
        capture_time: datetime,
        job_init_kwargs: dict,
        is_test: bool,
        transform_kwargs: dict,
        timestream_load_kwargs: dict,
    ) -> CustomTransformer:
        job_init_kwargs["aws_wrangler"] = self.__aws_wrangler
        job_init_kwargs["timestream_loader"] = self.__timestream_loader
        transformer: CustomTransformer = self.get_job(name, capture_time, job_init_kwargs)
        transformer.transform(is_test, transform_kwargs, timestream_load_kwargs)
        return transformer



import logging
from datetime import datetime, timedelta
from typing import Any, Optional

import pytz

from psi_collector import TIME_UNITS, get_logger
from psi_collector.custom_collector import CustomCollector


class CollectorRunner:
    def __init__(
        self,
        collectors: dict[str, Any],
        s3_loader: Optional[Any] = None,
        timestream_loader: Optional[Any] = None,
        logger: Optional[logging.Logger] = None,
    ):
        """
        Initializes a CollectorRunner.

        Accepts a collectors dict, mapping collector name to collector class.
        For example:
        {
            "divvy": DivvyCollector,
            "pagerduty": PagerDutyCollector
        }

        """
        assert collectors and isinstance(
            collectors, dict
        ), "A collectors dict is required to initialize a CollectorRunner"
        assert s3_loader, "A S3 loader is required to initialize a CollectorRunner"
        assert timestream_loader, "A Timestream loader is required to initialize a CollectorRunner"
        assert type(logger) in [
            logging.Logger,
            type(None),
        ], f"A logger of type logging.Logger is required to initialize a CollectorRunner. Got {logger}"
        self.__collectors = collectors
        self.__s3_loader = s3_loader
        self.__timestream_loader = timestream_loader
        self.__logger = logger or get_logger()

    def collect(self, event: dict):
        """Initializes and runs a collector.

        Accepts an event dictionary with the following required attributes:
            1. collector_name : the name of the collector to run
            2. s3_path        : the full s3 path to the data to collect from S3

        If writing to S3, the following attributes are required:
            1. write_to_s3    : boolean that must be set to true if writing to S3
            2. bucket_name    : S3 bucket to write to
            3. output_s3_path : S3 path to write to

        If writing to Timestream, the following attributes are required:
            1. write_to_timestream     : boolean that must be set to true if writing to Timestream
            2. database_name           : Timestream database to write to
            3. output_timestream_table : Timestream table to write to
            (optional) is_upsert       : whether or not to upsert Timestream records (default True)

        Optionally, capture_time may be set on the event as a millisecond timestamp. If unset,
        capture_time is taken from either the s3_path or the current time.
        """
        t0 = datetime.now()
        parsed_event = self.parse_event(event)
        self.__validate_parsed_event(parsed_event)
        self.__collect(**parsed_event)
        self.__logger.info(f"Collection completed in {str(datetime.now() - t0)}")

    def parse_event(self, event: dict) -> dict:
        """Parses a collector event into the following parts:
        1. collector_name         : the name of the collector to run
        2. capture_time           : time of collection
        3. write_to_s3            : boolean for whether or not to write to S3
        4. write_to_timestream    : boolean for whether or not to write to Timestream
        5. collector_init_kwargs  : kwargs passed into the collector's __init__
        6. collect_kwargs         : kwargs passed into the collector's collect function
        7. s3_load_kwargs         : kwargs passed into the S3 loader's load function
        8. timestream_load_kwargs : kwargs passed into the Timestream loader's load function

        Child classes may override this function to implement custom event parsing rules.
        """
        assert event and isinstance(
            event, dict
        ), f"Event must be a non-empty dictionary, got: {event}"
        return {
            "collector_name": event.get("collector_name"),
            "capture_time": self.get_capture_time(event),
            "write_to_s3": event.get("write_to_s3", True),
            "write_to_timestream": event.get("write_to_timestream", True),
            "collector_init_kwargs": {},
            "collect_kwargs": {
                "s3_path": event.get("s3_path"),
                "account_id": event.get("account_id"),
            },
            "s3_load_kwargs": {},
            "timestream_load_kwargs": {
                "upsert": event.get("is_upsert", True),
                "ignore_parse_error": True,
                "max_batch_size": event.get("max_batch_size"),
                "min_common_dimensions": event.get("min_common_dimensions"),
                "is_multi": bool(event.get("is_multi")),
            },
        }

    def __validate_parsed_event(self, parsed_event: dict):
        assert parsed_event and isinstance(
            parsed_event, dict
        ), f"Parsed event must be a non-empty dictionary, got: {parsed_event}"
        required_keys = [
            "collector_name",
            "capture_time",
            "write_to_s3",
            "write_to_timestream",
            "collector_init_kwargs",
            "collect_kwargs",
            "s3_load_kwargs",
            "timestream_load_kwargs",
        ]
        missing_keys = [k for k in required_keys if k not in parsed_event]
        extra_keys = [k for k in parsed_event if k not in required_keys]
        assert not missing_keys, f"Parsed event is missing keys: {missing_keys}"
        assert not extra_keys, f"Parsed event has extra keys: {extra_keys}"
        assert isinstance(
            parsed_event["collector_name"], str
        ), f"Invalid collector_name, expected a string but got: {parsed_event['collector_name']}"
        assert isinstance(
            parsed_event["capture_time"], datetime
        ), f"Invalid capture_time, expected a datetime but got: {parsed_event['capture_time']}"
        assert isinstance(
            parsed_event["write_to_s3"], bool
        ), f"Invalid write_to_s3, expected a bool but got: {parsed_event['write_to_s3']}"
        assert isinstance(
            parsed_event["write_to_timestream"], bool
        ), f"Invalid write_to_timestream, expected a bool but got: {parsed_event['write_to_timestream']}"
        assert isinstance(
            parsed_event["collector_init_kwargs"], dict
        ), f"Invalid collector_init_kwargs, expected a dict but got: {parsed_event['collector_init_kwargs']}"
        assert isinstance(
            parsed_event["collect_kwargs"], dict
        ), f"Invalid collect_kwargs, expected a dict but got: {parsed_event['collect_kwargs']}"
        assert isinstance(
            parsed_event["s3_load_kwargs"], dict
        ), f"Invalid s3_load_kwargs, expected a dict but got: {parsed_event['s3_load_kwargs']}"
        assert isinstance(
            parsed_event["timestream_load_kwargs"], dict
        ), f"Invalid timestream_load_kwargs, expected a dict but got: {parsed_event['timestream_load_kwargs']}"

        collect_kwargs = parsed_event["collect_kwargs"]
        if "s3_path" in collect_kwargs:
            assert type(collect_kwargs["s3_path"]) in [
                type(None),
                str,
            ], f"Invalid collect_kwargs.s3_path, expected a string but got: {collect_kwargs['s3_path']}"
        if "account_id" in collect_kwargs:
            assert type(collect_kwargs["account_id"]) in [
                type(None),
                str,
            ], f"Invalid collect_kwargs.account_id, expected a string but got: {collect_kwargs['account_id']}"

        timestream_load_kwargs = parsed_event["timestream_load_kwargs"]
        if "upsert" in timestream_load_kwargs:
            assert type(timestream_load_kwargs["upsert"]) in [
                type(None),
                bool,
            ], f"Invalid timestream_load_kwargs.upsert, expected a bool but got: {timestream_load_kwargs['upsert']}"
        if "ignore_parse_error" in timestream_load_kwargs:
            assert type(timestream_load_kwargs["ignore_parse_error"]) in [
                type(None),
                bool,
            ], f"Invalid timestream_load_kwargs.ignore_parse_error, expected a bool but got: {timestream_load_kwargs['ignore_parse_error']}"
        if "max_batch_size" in timestream_load_kwargs:
            assert type(timestream_load_kwargs["max_batch_size"]) in [
                type(None),
                int,
            ], f"Invalid timestream_load_kwargs.max_batch_size, expected an int but got: {timestream_load_kwargs['max_batch_size']}"
        if "min_common_dimensions" in timestream_load_kwargs:
            assert type(timestream_load_kwargs["min_common_dimensions"]) in [
                type(None),
                int,
            ], f"Invalid timestream_load_kwargs.min_common_dimensions, expected an int but got: {timestream_load_kwargs['min_common_dimensions']}"
        if "is_multi" in timestream_load_kwargs:
            assert type(timestream_load_kwargs["is_multi"]) in [
                type(None),
                bool,
            ], f"Invalid timestream_load_kwargs.is_multi, expected a bool but got: {timestream_load_kwargs['is_multi']}"

    def __collect(
        self,
        collector_name: str,
        capture_time: datetime,
        write_to_s3: bool,
        write_to_timestream: bool,
        collector_init_kwargs: dict,
        collect_kwargs: dict,
        s3_load_kwargs: dict,
        timestream_load_kwargs: dict,
    ):
        collector = self.get_collector(collector_name, capture_time, **collector_init_kwargs)
        data = collector.collect(**collect_kwargs)
        if write_to_s3:
            self.__logger.info("Beginning upload to S3")
            self.__s3_loader.load(collector_name, data, capture_time, **s3_load_kwargs)
        if write_to_timestream:
            self.__logger.info("Beginning write to Timestream")
            result = self.__timestream_loader.load(
                data=data,
                capture_time=capture_time,
                source=collector_name,
                **timestream_load_kwargs,
            )
            if len(result["invalid_items"]) > 0:
                raise Exception(
                    f"Failed to write to Timestream with invalid items {result['invalid_items']}"
                )
            if len(result["rejected_records"]) > 0:
                raise Exception(
                    f"Failed to write to Timestream with rejected records {result['rejected_records']}"
                )

    def get_capture_time(self, event: dict) -> datetime:
        if "capture_time" in event and event["capture_time"]:
            # fromisoformat returns a naive datetime object that IS NOT adjusted for the local timezone
            try:
                return datetime.fromisoformat(event["capture_time"]).replace(tzinfo=pytz.utc)
            except Exception:
                return self._parse_ago_time(event["capture_time"])
        elif "s3_path" in event and event["s3_path"]:
            capture_time_ms = event["s3_path"].split("/")[-1]
            # fromtimestamp returns a naive datetime object that IS adjusted for the local timezone
            return (
                datetime.fromtimestamp(int(capture_time_ms) / 1000)
                .astimezone()
                .astimezone(pytz.utc)
            )
        else:
            # now returns a naive datetime object that IS adjusted for the local timezone
            return datetime.now().astimezone().astimezone(pytz.utc)

    def _parse_ago_time(self, capture_time: str):
        try:
            if capture_time[-2:].isalpha():
                time_interval = TIME_UNITS[capture_time[-2:]]
                number = capture_time[0:-2]
            elif capture_time[-1:].isalpha():
                time_interval = TIME_UNITS[capture_time[-1:]]
                number = capture_time[0:-1]
            else:
                raise Exception(f"Unable to parse {capture_time}")
            timedelta_params = {time_interval: int(number)}
            return datetime.now() - timedelta(**timedelta_params)
        except Exception as e:
            self.__logger.error(f'Failed to parse ago time "{capture_time}" with exception: {e}')
            raise e

    def get_collector(
        self, collector_name: str, capture_time: datetime, **collector_init_kwargs
    ) -> CustomCollector:
        assert capture_time, "A capture_time is required to initialize a collector"
        cls = self.__collectors.get(collector_name)
        if cls:
            return cls(
                capture_time,
                collector_init_kwargs.pop("api", None),
                collector_init_kwargs.pop("aws_session", None),
                collector_init_kwargs.pop("http_session", None),
                **collector_init_kwargs,
            )
        raise Exception(f"Unrecognized collector with name: {collector_name}")



import logging
from datetime import datetime
from typing import Any, Optional

from typeguard import typechecked

from psi_elt.abstract_runner import AbstractRunner, NoneType
from psi_elt.custom_extractor import CustomExtractor
from psi_elt.utils import filter_dict


@typechecked
class ExtractorRunner(AbstractRunner):
    def __init__(
        self,
        extractors: dict[str, Any],
        s3_loader: Optional[Any] = None,
        logger: Optional[logging.Logger] = None,
    ):
        """
        Initializes a ExtractorRunner.

        Accepts a extractors dict, mapping extractor name to extractor class.
        For example:
        {
            "divvy": DivvyExtractor,
            "pagerduty": PagerDutyExtractor
        }
        """
        super().__init__(jobs=extractors, logger=logger)
        self.__s3_loader = s3_loader

    def parse_event(self, event: dict[str, Any]) -> dict[str, Any]:
        """Parses an extractor event. Child classes may override this function to
        implement custom event parsing rules.

        #### Required event keys
           1. name : the name of the extractor to run

        #### Optional event keys
           1. capture_time               : target time for the job to run against
           2. is_test                    : boolean for whethor or not this is a test run
           3. input_options.account_id   : AWS account ID to extract data from
           4. output_options.s3_uri      : S3 location to write data to
           5. output_options.s3_path     : S3 path to write data to
           6. output_options.write_to_s3 : boolean for whether or not to write to S3
        """
        parsed_event = {
            "name": event.get("name"),
            "capture_time": self.get_capture_time(event),
            "job_init_kwargs": {},
            "is_test": bool(event.get("is_test")),
            "extract_kwargs": {},
            "s3_load_kwargs": {},
        }

        # Parse input options
        parsed_event["extract_kwargs"] = filter_dict(
            dct=event.get("input_options"), keys=["account_id"]
        )

        # Parse output options
        parsed_event["s3_load_kwargs"] = filter_dict(
            dct=event.get("output_options"), keys=["s3_uri", "s3_path"]
        )
        parsed_event["extract_kwargs"].update(
            filter_dict(dct=event.get("output_options"), keys=["write_to_s3"])
        )
        return parsed_event

    def validate_parsed_event(self, parsed_event: dict[str, Any]):
        """Validates that a parsed event has the required format"""
        schema = {
            "name": str,
            "capture_time": datetime,
            "job_init_kwargs": {},
            "is_test": bool,
            "extract_kwargs": {"account_id": [str, NoneType], "write_to_s3": [bool, NoneType]},
            "s3_load_kwargs": {"s3_uri": [str, NoneType], "s3_path": [str, NoneType]},
        }
        try:
            self._validate_dict(parsed_event, schema)
        except AssertionError as e:
            self.logger.error("Invalid event")
            raise e

    def _run(
        self,
        name: str,
        capture_time: datetime,
        job_init_kwargs: dict,
        is_test: bool,
        extract_kwargs: dict,
        s3_load_kwargs: dict,
    ) -> CustomExtractor:
        job_init_kwargs["s3_loader"] = self.__s3_loader
        extractor: CustomExtractor = self.get_job(name, capture_time, job_init_kwargs)
        extractor.extract(is_test, extract_kwargs, s3_load_kwargs)
        return extractor
